@@ -6,7 +6,7 @@ use proc_macro2::{Ident, Span, TokenStream};
 use quote::TokenStreamExt;
 
 pub mod attributes {
-    use crate::ir::comp::SpecialMemberKind;
+    use crate::ir::{comp::SpecialMemberKind, function::Visibility};
     use proc_macro2::{Ident, Span, TokenStream};
     use std::str::FromStr;
 
@@ -94,6 +94,19 @@ pub mod attributes {
         }
     }
 
+    pub fn visibility(visibility: Visibility) -> TokenStream {
+        match visibility {
+            Visibility::Public => quote! {
+            },
+            Visibility::Protected => quote! {
+                #[bindgen_visibility_protected]
+            },
+            Visibility::Private => quote! {
+                #[bindgen_visibility_private]
+            },
+        }
+    }
+
     pub fn unused_template_param_in_arg_or_return() -> TokenStream {
         quote! {
             #[bindgen_unused_template_param_in_arg_or_return]
